using System.Collections.Generic;
using System.Drawing;
using System.IO;

using VirtualLab.Programming;

using VirtualLabAPI.Core.BasicFunctions;
using VirtualLabAPI.Core.Common;
using VirtualLabAPI.Core.FieldRepresentations;
using VirtualLabAPI.Core.Numerics;
using VirtualLabAPI.Core.ParameterRuns;
using VirtualLabAPI.Core.ExportImport;

namespace OwnCode {
	/// <summary>
	/// Module for exporting the results of a Parameter Run to hard disc.
	/// </summary>
	/// <remarks>
	/// • Author: LightTrans International GmbH and Wyrowski Photonics GmbH
	/// • Version: 3.0α, Date: 2022-03-17
	/// • Tested with VirtualLab 2021.1.
	/// • License: CC-BY 3.0
	/// </remarks>
	public class VLModule {
		//enter the indices of the desired detectors: 
		int indexCameraDetector = 611;
		int indexValueDetector = 612;
		string subdetectorValue = "RMS [λ] of Wavefront Error";

		//enter the desired directory for export:
		string directory = @"C:\Temp\Results\";

		/// <summary>
		/// Main function
		/// </summary>
		public void Run() {

			ParameterRun parameterRun = DocumentHistory<ParameterRun>.BrowseDocuments("Select Parameter Run for Export", returnReference: true);

			if (parameterRun == null) {
				VL_GUI.WriteLineToMessagesTab("Export was canceled.");
				return;
			}

			if (!Directory.Exists(directory)) {
				Directory.CreateDirectory(directory);
			}

			ExportDocuments(parameterRun, indexCameraDetector, $"Detector {indexCameraDetector} - Document ");
			ExportChromaticFieldsSetsAsBitmaps(parameterRun, indexCameraDetector, $"Detector {indexCameraDetector} - Bitmap ");
			ExportPhysicalValuesAsTextFile(parameterRun, indexValueDetector, subdetectorValue, $"Detector {indexValueDetector} - Text");
		}

		/// <summary>
		/// Saves the documents generated by a specific detector.
		/// </summary>
		/// <param name="prToExport">The Parameter Run from which the data is exported.</param>
		/// <param name="detectorIndex">The index of the detector whose data is exported.</param>
		/// <param name="fileNamePrefix">The prefix added to all exported files.</param>
		public void ExportDocuments(ParameterRun prToExport, int detectorIndex, string fileNamePrefix) {
			getMatchingResults(prToExport, detectorIndex, out List<PhysicalValue> iterationHeaders, out List<IElementaryParameterRunResult> listResults);

			//loop over all results
			for (int runResults = 0; runResults < listResults.Count; runResults++) {
				//extract current result
				IElementaryParameterRunResult currResult = listResults[runResults];

				//if result is not available, e.g. for only partially calculated Parameter Run ⇒ skip
				if (currResult == null) {
					continue;
				}

				//convert as document
				IDocument resultAsDocument = currResult.ActualParameterRunResult as IDocument;

				//generate filename
				int numberOfFile = runResults + 1;
				string fileName = Path.Combine(directory, $"{fileNamePrefix}{numberOfFile:0000}.{resultAsDocument.FileExtension}");

				resultAsDocument.Save(fileName);
				Globals.DataDisplay.LogMessage($"Export to “{fileName}” was successful.");
			}
		}

		/// <summary>
		/// Exports the Chromatic Fields Sets of a specific detector as bitmap files.
		/// </summary>
		/// <param name="prToExport">The Parameter Run from which the data is exported.</param>
		/// <param name="detectorIndex">The index of the detector whose data is exported.</param>
		/// <param name="fileNamePrefix">The prefix added to all exported files.</param>
		public void ExportChromaticFieldsSetsAsBitmaps(ParameterRun prToExport, int detectorIndex, string fileNamePrefix) {
			getMatchingResults(prToExport, detectorIndex, out List<PhysicalValue> iterationHeaders, out List<IElementaryParameterRunResult> listResults);

			//loop over all results
			for (int runResults = 0; runResults < listResults.Count; runResults++) {
				//extract current result
				IElementaryParameterRunResult currResult = listResults[runResults];

				//if result is not available, e.g. for only partially calculated Parameter Run ⇒ skip
				if (currResult == null) {
					continue;
				}

				//convert as chromatic fields set
				ChromaticFieldsSet resultAsCFS = currResult.ActualParameterRunResult as ChromaticFieldsSet;

				DataArray2D daSumOfAllSubsets = resultAsCFS.GetSumOfAllSubsets() as DataArray2D;

				//generate complex amplitude from given complex field
				ComplexAmplitude caField = new ComplexAmplitude(daSumOfAllSubsets.Data[0]);
				//set correct sampling parameters
				caField.SamplingDistance = new VectorD(daSumOfAllSubsets.SamplingDistance_X, daSumOfAllSubsets.SamplingDistance_Y);

				double min, max;
				//get min/max
				ComplexFieldEvaluation.GetMinMaxValues(daSumOfAllSubsets.Data[0],
													   FieldQuantity.RealPart,
													   out min,
													   out max);

				//generate colormapping object
				ColorMapping cMapping = new ColorMapping();
				cMapping.DefineFieldQuantity(FieldQuantity.RealPart, min, max, startColor: Color.Black, endColor: Color.White, colorFromWavelength: false);

				//generate filename
				int numberOfFile = runResults + 1;
				string currFileNameBMP = Path.Combine(directory, $"{fileNamePrefix}{numberOfFile:0000}.bmp");

				//perform export to BMP
				ImportExport.ExportBitmap(caField,
										  currFileNameBMP,
										  cMapping);
			}
		}

		/// <summary>
		/// Exports the physical values of one specific detector and sub-detector into a single text file.
		/// </summary>
		/// <param name="prToExport">The Parameter Run from which the data is exported.</param>
		/// <param name="detectorIndex">The index of the detector whose data is exported.</param>
		/// <param name="subDetector">The sub-detector whose values are exported.</param>
		/// <param name="fileName">The file name (without extension) to which the values are exported.</param>
		public void ExportPhysicalValuesAsTextFile(ParameterRun prToExport, int detectorIndex, string subDetector, string fileName) {
			getMatchingResults(prToExport, detectorIndex,
							   out List<PhysicalValue> iterationHeaders, out List<IElementaryParameterRunResult> listResults,
							   subDetector);

			List<double> values = new List<double>();

			TextExportImportSettings exportSettings = new TextExportImportSettings();
			exportSettings.ColumnSeparator = '\t';
			exportSettings.DecimalSeparator = '.';
			exportSettings.HeaderIndicator = '#';

			//generate header
			string header = "# Origin: Variation of " + iterationHeaders[0].Comment + "\r\n";

			PhysicalValue firstValue = listResults[0].ActualParameterRunResult as PhysicalValue;
			header += $"# Description: {firstValue.Comment}\r\n";
			header += $"# Physical Property: {EnumDescription.GetDescription(firstValue.PhysicalProperty)}\r\n";

			//loop over all results for both Detectors
			for (int runResults = 0; runResults < listResults.Count; runResults++) {
				//extract current result
				IElementaryParameterRunResult currResult = listResults[runResults];

				//if result is not available, e.g. for only partially calculated Parameter Run ⇒ skip
				if (currResult == null) {
					continue;
				}

				values.Add((currResult.ActualParameterRunResult as PhysicalValue).Value);
			}

			// Create final complex field
			ComplexField complexField = new ComplexField(new Vector(1, values.Count), isComplex: false, PrecisionMode.Double);
			for (int i = 0; i < values.Count; i++) {
				complexField[0, i] = values[i];
			}

			//perform txt export
			string completeFileName = Path.Combine(directory, $"{fileName}.txt");
			ImportExportNumerics.ExportComplexField(complexField, header, FieldQuantity.RealPart, 1, '.', '\t', completeFileName);
		}

		/// <summary>
		/// Support method to get all results of a given detector (and sub-detector).
		/// </summary>
		/// <param name="prToExport">The Parameter Run from which the data is exported.</param>
		/// <param name="detectorIndex">The index of the detector whose data is exported.</param>
		/// <param name="iterationHeaders">Returns the column headers for all results.</param>
		/// <param name="listResults">Returns the actual results for the given detector (and sub-detector).</param>
		/// <param name="subDetector">Optional sub-detector. Usually required for a detector returning physical values.</param>
		private void getMatchingResults(ParameterRun prToExport, int detectorIndex,
										out List<PhysicalValue> iterationHeaders, out List<IElementaryParameterRunResult> listResults,
										string subDetector = "") {
			string shortDetectorName = prToExport.LightPathToShow[detectorIndex].CompleteTitle;
			string longDetectorName = "";

			#region search within the stored detector names (which contain also info about the linakge and simulation engine)
			//search for index of result
			for (int runSearchIndex = 0; runSearchIndex < prToExport.ResultMatrix.NameOfResults.Count; runSearchIndex++) {
				//get current string of result
				string currResultCaption = prToExport.ResultMatrix.NameOfResults[runSearchIndex];

				//check whether index of camera detector is within caption => then stop and store index
				if (currResultCaption.Contains(shortDetectorName)) {
					//store index for export
					longDetectorName = currResultCaption;
					//stop loop
					break;
				}
			}
			#endregion

			if (longDetectorName.Equals("")) {
				Globals.DataDisplay.LogError($"Detector with index {detectorIndex} could not be found!");
			}

			prToExport.ResultMatrix.GetResultsOfSpecificDetector(longDetectorName, subDetector, 0, prToExport.ResultMatrix.NumberOfIterations - 1, out iterationHeaders, out listResults);
		}
	}
}
